import unittest
import yaml

from tests.demo_config import Config


class Yaml2ConfigTest(unittest.TestCase):
    def setUp(self) -> None:
        self.path_config = "tests/config.yaml"

    def test(self):
        data_class = Config.from_yaml(self.path_config)
        with open(self.path_config, "r") as file:
            data_yaml = yaml.load(file, Loader=yaml.FullLoader)
        self.assertTrue(self.compare_yaml_dict_and_generated_class(data_yaml, data_class))

    def compare_yaml_dict_and_generated_class(self, data_yaml, data_class) -> bool:
        equal = True
        for key, val_yaml in data_yaml.items():
            val_class = data_class.__getattribute__(key)
            if isinstance(val_yaml, dict):
                equal = equal and self.compare_yaml_dict_and_generated_class(val_yaml, val_class)
            else:
                equal = equal and (val_yaml == val_class)
        return equal



import io
import os

import setuptools

# Package meta-data.
NAME = "yaml2pyclass"
DESCRIPTION = "Code generator that produces a Python class from a YAML input file. " \
              "Can be used to facilitate code completion for config objects. "
LONG_DESCRIPTION_CONTENT_TYPE = "text/markdown"
URL = "https://github.com/MaximChernyak98/yaml2pyclass-mecher"
EMAIL = "alex.code@mail.com, hertlein@fzi.de"
AUTHOR = "Alexander Naumann, Felix Hertlein"
REQUIRES_PYTHON = ">=3.7.0"

# Package requirements.
base_packages = ["pyyaml==6.0"]

# dev_packages = base_packages + [
#     "black>=20.1.0",
#     "isort==5.9.3",
#     "mypy>=0.761",
#     "pre-commit>=2.9.2",
#     "pytest>=4.5.0",
#     "pytest-cov>=2.6.1",
#     "pyupgrade>=3.2.0",
#     "types-PyYAML",
# ]

# docs_packages = [
#     "flask>=2.0.2",
#     "ipykernel>=6.9.0",
#     "mike>=0.5.3",
#     "mkdocs>=1.2.3",
#     "mkdocs-awesome-pages-plugin>=2.7.0",
#     "mkdocs-gen-files>=0.3.5",
#     "mkdocs-literate-nav>=0.4.1",
#     "mkdocs-material>=8.1.11",
#     "mkdocstrings[python]>=0.19.0",
#     "pytkdocs[numpy-style]>=0.5.0",
#     "ipython_genutils>=0.1.0",
#     "mkdocs-jupyter>=0.20.0",
#     "mkdocs-bibtex==2.8.1",
#     "nbconvert==6.4.2",
#     "numpydoc==1.2",
#     "spacy==3.2.2",
#     "jinja2==3.0.3",
# ]

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = "\n" + f.read()

# Load the package's __version__.py module as a dictionary.
about = {}
with open(os.path.join(here, NAME, "__version__.py")) as f:
    exec(f.read(), about)

# Where the magic happens:
setuptools.setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=setuptools.find_packages(exclude=("tests",)),
    install_requires=base_packages,
    # extras_require={
    #     "dev": dev_packages,
    #     "test": dev_packages,
    #     "docs": docs_packages,
    #     "all": dev_packages + docs_packages,
    # },
    include_package_data=True,
    license="MIT",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    ext_modules=[],
)



import inspect
import os
import re
from dataclasses import dataclass
from inspect import getmembers
from pathlib import Path
from typing import Any, Dict, List

import yaml


@dataclass
class CodeGenerator:
    @classmethod
    def from_yaml(cls, config_file: str) -> "CodeGenerator":
        if not os.path.exists(config_file):
            raise FileNotFoundError(f"Could not find YAML file at {config_file}")
        with open(config_file, "r") as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
            data = {} if data is None else data

        cls._update_code(data)

        old_fields = list(
            [value for key, value in getmembers(cls) if key == "__dataclass_fields__"][
                0
            ].keys()
        )
        dummy_data = {field: None for field in old_fields}

        c = cls(**dummy_data)
        cls._update_data(c, data)

        return c

    @classmethod
    def _update_data(cls, obj: object, data: Dict[str, Any]) -> object:

        obj.__dict__.clear()
        obj.__dict__.update(data)

        for key, value in data.items():
            if not isinstance(value, dict):
                continue

            subclass = type(cls._to_class_name(key), (), {})
            obj.__dict__[key] = cls._update_data(subclass(), value)

        return obj

    @classmethod
    def _update_code(cls, data: Dict[str, Any]) -> None:
        code_file = Path(inspect.getfile(cls))
        lines = code_file.read_text().split("\n")

        start_pos_list = [
            idx
            for idx, line in enumerate(lines)
            if ("class" in line and cls.__name__ in line and "Class" not in line)
        ]
        if len(start_pos_list) != 1:
            raise ValueError("Malformed config file! Cannot modify file!")

        # find indentation
        pos = start_pos_list[0] + 1
        match = re.match(pattern=r"^(?P<indentation>\s+)", string=lines[pos])
        if match is None:
            raise ValueError("Malformed config file! Cannot modify file!")
        indentation = match.groupdict()["indentation"]

        # clear old keys
        while pos < len(lines):
            lines.pop(pos)

        # add new keys
        lines[pos:pos] = [
            indentation + line for line in cls._create_code(data, indentation)
        ] + [""]

        # add import
        cls._add_import(lines)

        code_file.write_text("\n".join(lines))

    @classmethod
    def _add_import(cls, lines: List[str]) -> None:
        pattern = r"^\s*import\s*dataclasses\s*$"
        if any(re.match(pattern=pattern, string=line) for line in lines):
            return

        lines.insert(0, "import dataclasses")

    @classmethod
    def _create_code(cls, data: Dict[str, Any], indentation: str) -> List[str]:
        if len(data.keys()) == 0:
            return ["pass"]

        dict_keys = [key for key, value in data.items() if isinstance(value, dict)]

        lines = []
        for key in dict_keys:
            lines.append("@dataclasses.dataclass")
            lines.append(f"class {cls._to_class_name(key)}:")
            lines.extend(
                [
                    indentation + line
                    for line in cls._create_code(data[key], indentation)
                ]
            )
            lines.append("")

        for key, value in data.items():
            if key in dict_keys:
                lines.append(f"{str(key)}: {cls._to_class_name(key)}")
            else:
                if isinstance(value, str) and 'type___' in value:
                    new_type = value.split('type___')[1]
                    lines.append(f"{str(key)}: {new_type}")
                else:
                    lines.append(f"{str(key)}: {type(value).__name__}")

        return lines

    @staticmethod
    def _to_class_name(key: str) -> str:
        return "".join(x.title() for x in key.split("_")) + "Class"
